##############################################################################
#
# Copyright (c) 2004 TINY SPRL. (http://tiny.be) All Rights Reserved.
#                    Fabien Pinckaers <fp@tiny.Be>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

#
# OSV: Objects Services
#

import orm
import fields as cols
import netsvc
import psycopg
import sql_db

class except_osv(Exception):
	def __init__(self, name, value, exc_type='warning'):
		self.name = name
		self.exc_type = exc_type
		self.value = value
		self.args = (exc_type,name)

class osv_pool(netsvc.Service):
	obj_pool = {}
	def __init__(self):
		netsvc.Service.__init__(self, 'object_proxy', audience='')
		self.joinGroup('web-services')
		self.exportMethod(self.exportedMethods)
		self.exportMethod(self.obj_list)
		self.exportMethod(self.exec_workflow)
		self.exportMethod(self.execute)
		self.exportMethod(self.execute_cr)

	def execute_cr(self, cr, uid, obj, method, *args, **kw):
		#
		# TODO: check security level
		#
		try:
			if (not method in getattr(self.obj_pool[obj],'_protected')) and len(args) and args[0] and len(self.obj_pool[obj]._inherits):
				types = {obj: args[0]}
				cr.execute('select inst_type,inst_id,obj_id from inherit where obj_type=%s and  obj_id in ('+','.join(map(str,args[0]))+')', (obj,))
				for ty,id,id2 in cr.fetchall():
					if not ty in types:
						types[ty]=[]
					types[ty].append(id)
					types[obj].remove(id2)
				for t,ids in types.items():
					if len(ids):
						t = self.obj_pool[t]
						res = getattr(t,method)(cr, uid, ids, *args[1:], **kw)
			else:
				obj = self.obj_pool[obj]
#				print 'obj', obj
				res = getattr(obj,method)(cr, uid, *args, **kw)
			return res
		except orm.except_orm, inst:
			self.abortResponse(1, inst.value[0], inst.name, inst.value[1])
		except except_osv, inst:
			self.abortResponse(1, inst.name, inst.exc_type, inst.value)

	def execute(self, uid, obj, method, *args, **kw):
#		print 'executing', method,args
		cr=sql_db.db.cursor()
		res = self.execute_cr(cr, uid, obj, method, *args, **kw)
		cr.commit()
		cr.close()
		return res
#		except psycopg.IntegrityError:
#			raise 'SQL Fault'

	def exec_workflow_cr(self, cr, uid, obj, method, *args):
#		try:
			wf_service = netsvc.LocalService("workflow")
			wf_service.trg_validate(uid, obj, args[0], method, cr)
			return True
#		except orm.except_orm, inst:
#			print 'abort'
#			self.abortResponse(1, inst.value[0], 'ValidateError', inst.value[1])

	def exec_workflow(self, uid, obj, method, *args):
#		try:
			cr=sql_db.db.cursor()
			res = self.exec_workflow_cr(cr, uid, obj, method, *args)
			cr.commit()
			cr.close()
			return res
#		except Exception, e:
#			print e.args
#			print e[0]
#			print dir(e)
#			print type(e)
#		except netsvc.ServiceFailure, inst:
#			self.abortResponse(inst.error, inst.description, inst.origin, inst.details)

	def obj_list(self):
		return self.obj_pool.keys()

	# adds a new object instance to the object pool. 
	# if it already existed, the instance is replaced
	def add(self, name, obj_inst):
		if self.obj_pool.has_key(name):
#TODO: print warning to logger.debug
			print "replacing object", name 
			del self.obj_pool[name]
		self.obj_pool[name] = obj_inst

	def get(self, name):
		return self.obj_pool.get(name, None)

osv_pools = osv_pool()

class inheritor(type):
	def __new__(cls, name, bases, d):
		parent_name = d.get('_inherit', None)
		if parent_name:
			parent_obj = osv_pools.get(parent_name)
			assert parent_obj, "parent object %s does not exist !" % parent_name
			for s in ('_columns', '_defaults'):
				new_dict = getattr(parent_obj.__class__, s)
				new_dict.update(d.get(s, {}))
				d[s] = new_dict
			bases = (parent_obj.__class__,)
		return type.__new__(cls, name, bases, d)

class osv(orm.orm):
	__metaclass__ = inheritor

	def __init__(self):
		osv_pools.add(self._name, self)
		self.pool = osv_pools
		orm.orm.__init__(self)

class Cacheable(object):

	_cache = {}
	_items = []

	def add(self, model, id, context, values):
		print 'Caching:', model, id, context
		self._items.append((model, id, context))
		if len(self._cache) > 1000:
			m, i , c = self._items[0]
			del self._items[0]
			del self._cache[model][id][context]
		self._cache.setdefault(model, {}).setdefault(id, {})[context] = values
	
	def invalidate(self, model, id):
		for idx, t in enumerate(self._items[:]):
			if t[0] == model and t[1] == id:
				del self._items[idx]
		del self._cache[model][id]
	
	def get(self, model, id, context):
		try:
			return self._cache[model][id][context]
		except KeyError:
			return None
	
	def clear(self):
		self._cache.clear()

def filter_dict(d, fields):
	res = {}
	for f in fields + ['id']:
		if f in d:
			res[f] = d[f]
	return res

def hashablify(context, t):
	if t is dict:
		ctx = context.items()
		for idx, item in enumerate(ctx):
			key, value = item
			if isinstance(value, (list, dict)):
				ctx[idx] = (key, hashablify(value, type(value)))
	elif t is list:
		ctx = context[:]
		for idx, item in enumerate(ctx):
			if isinstance(item, (list, dict)):
				ctx[idx] = hashablify(item, type(item))
	else:
		return hash(context)

	ctx.sort()
	return tuple(ctx)

cacheable_osv = osv

class cacheable_osv_bak(osv, Cacheable):

	def __init__(self):
		super(cacheable_osv, self).__init__()
	
	def read(self, cr, user, ids, fields=[], context={}, load='_classic'):
		fields = fields or self._columns.keys()
		ctx = hashablify(context, type(context))
		result, tofetch = [], []
		for id in ids:
			res = self.get(self._name, id, ctx)
			if not res:
				tofetch.append(id)
			else:
				result.append(filter_dict(res, fields))

		# Il faudrait faire attention aux one2many, car ils contiennent une
		# reference vers l'objet que l'on lit.
		nfields = [key for key, col in self._columns.items() if not col._classic ]

		res = super(cacheable_osv, self).read(cr, user, tofetch, nfields, context, load)
		for r in res:
			self.add(self._name, r['id'], ctx, r)
			result.append(filter_dict(r, fields))

		# Appel de fonction si necessaire
		tofetch = []
		for f in fields:
			if f not in nfields:
				tofetch.append(f)
		for f in tofetch:
			fvals = self._columns[f].get(cr, self, ids, f, user, context=context)
			for r in result:
				r[f] = fvals[r['id']]

		# TODO: tri par self._order !!
		return result

	def write(self, cr, user, ids, values, context={}):
		for id in ids:
			self.invalidate(self._name, id)
		return super(cacheable_osv, self).write(cr, user, ids, values, context)
	
	def unlink(self, cr, user, ids):
		self.clear()
		return super(cacheable_osv, self).unlink(cr, user, ids)

# vim:noexpandtab:
