##############################################################################
#
# Copyright (c) 2004 TINY SPRL. (http://tiny.be) All Rights Reserved.
#                    Fabien Pinckaers <fp@tiny.Be>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################


import os, time
import ir, netsvc, sql_db
import osv
from interface import  report_rml
import re

import copy

parents = {
	'tr':1,
	'li':1,
	'story': 0,
	'section': 0
}


#
# Context: {'node': node.dom}
# 
class browse_record_list(list):
	def __init__(self, lst, context):
		super(browse_record_list, self).__init__(lst)
		self.parents = copy.copy(parents)
		self.context = context

	def __getattr__(self, name):
		res = browse_record_list([getattr(x,name) for x in self], self.context)
		return res

	def __str__(self):
		return "browse_record_list("+str(len(self))+")"

	def repeatIn(self, name):
		node = self.context['_node']
		node.data = ''
		while True:
			if not node.parentNode:
				break
			node = node.parentNode
			if node.nodeType == node.ELEMENT_NODE and node.localName in parents:
				break
		parent_node = node
		if not len(self):
			return None
		nodes = [(0,node)]
		for i in range(1,len(self)):
			newnode = parent_node.cloneNode(1)
			n = parent_node.parentNode
			n.insertBefore(newnode, parent_node)
			nodes.append((i,newnode))
		for i,node in nodes:
			self.context[name] = self[i]
			self.context['_self']._parse_node(node)
		return None

class rml_parse(object):
	def __init__(self):
		self.regex = re.compile('\[\[(.+?)\]\]')
		self.context = {}

	def _eval(self, expr, node):
		self.context['_node'] = node
		self.context['objects'] = self.objs
		self.context['_self'] = self
		res = eval(expr, self.context) or ''
		return res

	def _find_parent(self, node):
		while True:
			if not node.parentNode:
				return False
			node = node.parentNode
			if node.nodeType == node.ELEMENT_NODE and node.localName in parents:
				break
		return node

	def _parse_text(self, node, text, level=[]):
		res = self.regex.findall(text)
		todo = []
		for key in res:
			newtext = self._eval(key, node)
			for i in range(len(level)):
				if isinstance(newtext,browse_record_list) or (type(newtext)==type([])):
					newtext = newtext[level[i]]
			if isinstance(newtext,browse_record_list) or (type(newtext)==type([])):
				todo.append((key, newtext))
			else:
				if type(newtext) not in (type(''), type(u'')):
					newtext = str(newtext or '')
				txt = text.replace('[['+key+']]', newtext.decode('utf8'))
				node.data = txt
		if len(todo):
			for key, newtext in todo:
				parent_node = self._find_parent(node)
				assert parents.get(parent_node.localName,False), 'No parent node found !'
				nodes = [parent_node]
				for i in range(len(newtext)-1):
					newnode = parent_node.cloneNode(1)
					if parents.get(parent_node.localName,False):
						n = parent_node.parentNode
						parent_node.parentNode.insertBefore(newnode, parent_node)
						nodes.append(newnode)
				for n in range(len(nodes)):
					self._parse_node(nodes[n], level + [n])
			return False
		return text

	def _parse_node(self, node, level=[]):
		for n in node.childNodes:
			if n.nodeType == n.ELEMENT_NODE:
				self._parse_node(n, level)
			elif (n.nodeType == node.CDATA_SECTION_NODE) or (n.nodeType == node.TEXT_NODE):
				self._parse_text(n,n.data,level)

	def _parse(self, rml, objs, context):
		import xml.dom.minidom
		self.context = context
		self.objs = objs
		self.dom = xml.dom.minidom.parseString(rml)
		el = self.dom.documentElement
		self._parse_node(el)
		return el.toxml('utf-8')


class report_sxw(report_rml):
	def __init__(self, name, table, rml):
		report_rml.__init__(self, name, table, rml, '')
		self.bin_datas = {}

	def create(self, uid, ids, datas, context={}):
		rml = file(self.tmpl).read()

		cr = sql_db.db.cursor()
		objs = osv.osv.osv_pools.get(self.table).browse(cr, uid, ids, list_class=browse_record_list, context=context)
		rml_parser = rml_parse()
		rml2 = rml_parser._parse(rml, objs, context)
		cr.close()

		create_doc = self.generators[datas.get('report_type','pdf')]
		pdf = create_doc(rml2)
		return pdf 

